cimport numpy as np
from numpy.math cimport NAN

import numpy as np

np.import_array()

cdef extern from "stats_stubs.c":
    ctypedef struct stats_t:
        np.int64_t count
        np.float64_t sum
        np.float64_t min
        np.float64_t max
        np.float64_t m2
        np.float64_t m3
        np.float64_t m4
        bint homogeneous
        np.float64_t first_value


    stats_t *stats_new()
    void stats_free(stats_t *T)
    void stats_add(stats_t *T, np.float64_t x, np.int64_t count)
    void stats_merge(stats_t *T1, stats_t *T2)
    np.intp_t stats_update_ndarray(stats_t *T, np.PyArrayObject *x,
                                   np.PyArrayObject *w) except -1
    np.float64_t stats_mean(stats_t *T)
    np.float64_t stats_var(stats_t *T, np.int64_t ddof)
    np.float64_t stats_std(stats_t *T, np.int64_t ddof)
    np.float64_t stats_skew(stats_t *T, int bias)
    np.float64_t stats_kurt(stats_t *T, int fischer, int bias)


cdef class SummaryStats:
    """SummaryStats()

    Computes exact summary statistics on a data stream.

    Keeps track of enough information to compute:

    - count
    - sum
    - minimum
    - maximum
    - mean
    - variance
    - standard deviation
    - skewness
    - kurtosis

    Notes
    -----
    The update formulas for variance, standard deviation, skewness, and
    kurtosis were taken from [1]_.

    References
    ----------
    .. [1] Pebay, Philippe. "Formulas for robust, one-pass parallel computation
       of covariances and arbitrary-order statistical moments." Sandia Report
       SAND2008-6212, Sandia National Laboratories 94 (2008).
    """
    cdef stats_t *stats

    def __cinit__(self):
        self.stats = stats_new()

    def __dealloc__(self):
        if self.stats != NULL:
            stats_free(self.stats)

    def __repr__(self):
        return "SummaryStats<count=%d>" % self.count()

    def __reduce__(self):
        return (SummaryStats, (), self.__getstate__())

    def __getstate__(self):
        return (self.stats.count, self.stats.sum, self.stats.min,
                self.stats.max, self.stats.m2, self.stats.m3, self.stats.m4, 
                self.stats.homogeneous, self.stats.first_value)

    def __setstate__(self, state):
        self.stats.count = state[0]
        self.stats.sum = state[1]
        self.stats.min = state[2]
        self.stats.max = state[3]
        self.stats.m2 = state[4]
        self.stats.m3 = state[5]
        self.stats.m4 = state[6]
        self.stats.homogeneous = state[7]
        self.stats.first_value = state[8]

    def add(self, double x, int count=1):
        """add(self, x)

        Add an element to the summary.

        Parameters
        ----------
        x : double
        count : int, optional
            The count of the item to add. Default is 1.
        """
        if count <= 0:
            raise ValueError("count must be > 0")
        stats_add(self.stats, x, count)

    def update(self, x, count=1):
        """update(self, x)

        Add many elements to the summary.

        Parameters
        ----------
        x : array_like
        count : array_like, optional
            The count (or counts) of the item to add. Default is 1.
        """
        x = np.asarray(x)
        check_count = not (np.isscalar(count) and count == 1)
        count = np.asarray(count, dtype='i8')
        if check_count and (count <= 0).any():
            raise ValueError("count must be > 0")

        stats_update_ndarray(self.stats, <np.PyArrayObject*>x,
                             <np.PyArrayObject*>count)

    def merge(self, *args):
        """merge(self, *args)

        Update this summary inplace with data from other summaries.

        Parameters
        ----------
        args : SummaryStats
            SummaryStats to merge into this one.
        """
        if not all(isinstance(i, SummaryStats) for i in args):
            raise TypeError("All arguments to merge must be SummaryStats")
        cdef SummaryStats s
        for s in args:
            stats_merge(self.stats, s.stats)

    def count(self):
        """count(self)

        The number of elements in the summary.
        """
        return self.stats.count

    def sum(self):
        """sum(self)

        The sum of all elements in the summary.
        """
        return self.stats.sum

    def min(self):
        """min(self)

        The minimum element added to the summary. Returns ``NaN`` if empty.
        """
        return self.stats.min if self.stats.count > 0 else NAN

    def max(self):
        """max(self)

        The maximum element added to the summary. Returns ``NaN`` if empty.
        """
        return self.stats.max if self.stats.count > 0 else NAN

    def mean(self):
        """mean(self)

        The mean of all elements in the summary. Returns ``NaN`` if empty.
        """
        return stats_mean(self.stats)

    def var(self, int ddof=0):
        """var(self, ddof=0)

        The variance of all elements in the summary. Returns ``NaN`` if empty.

        Parameters
        ----------
        ddof : int, optional
            "Delta Degrees of Freedom": the divisor used in the calculation is
            ``N - ddof``, where ``N`` represents the number of elements. By
            default ``ddof`` is zero.
        """
        return stats_var(self.stats, ddof)

    def std(self, int ddof=0):
        """std(self, ddof=0)

        The standard deviation  of all elements in the summary. Returns
        ``NaN`` if empty.

        Parameters
        ----------
        ddof : int, optional
            "Delta Degrees of Freedom": the divisor used in the calculation is
            ``N - ddof``, where ``N`` represents the number of elements. By
            default ``ddof`` is zero.
        """
        return stats_std(self.stats, ddof)

    def skew(self, bint bias=True):
        """skew(self, bias=True)

        The skewness of all elements in the summary. Returns ``NaN`` if empty.

        Parameters
        ----------
        bias : bool, optional
            If False, then the calculations are corrected for statistical bias.
            Default is True.
        """
        return stats_skew(self.stats, bias)

    def kurt(self, bint fisher=True, bint bias=True):
        """kurt(self, fisher=True, bias=True)

        The kurtosis (Fisher or Pearson) of all elements in the summary.
        Returns ``NaN`` if empty.

        Parameters
        ----------
        fisher : bool, optional
            If True [default], Fisher's definition is used (normal ==> 0.0). If
            False, Pearson's definition is used (normal ==> 3.0).
        bias : bool, optional
            If False, then the calculations are corrected for statistical bias.
            Default is True.
        """
        return stats_kurt(self.stats, fisher, bias)



from __future__ import absolute_import

from ._version import get_versions
from .space_saving import SpaceSaving
from .stats import SummaryStats
from .tdigest import TDigest

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



from libc.string cimport memcpy
from numpy.math cimport NAN, isfinite

from copy import copy

cimport cython
cimport numpy as np
import numpy as np

np.import_array()


cdef extern from "tdigest_stubs.c":
    ctypedef struct centroid_t:
        double mean
        double weight

    # Not the full struct, just the parameters we want
    ctypedef struct tdigest_t:
        double compression
        size_t size
        size_t buffer_size

        double min
        double max

        size_t ncentroids
        double total_weight
        double buffer_total_weight
        centroid_t *centroids

    tdigest_t *tdigest_new(double compression)
    void tdigest_free(tdigest_t *T)
    void tdigest_add(tdigest_t *T, double x, double w)
    void tdigest_flush(tdigest_t *T)
    void tdigest_merge(tdigest_t *T, tdigest_t *other)
    void tdigest_scale(tdigest_t *T, double factor)
    np.npy_intp tdigest_update_ndarray(tdigest_t *T, np.PyArrayObject *x, np.PyArrayObject *w) except -1
    np.PyArrayObject *tdigest_quantile_ndarray(tdigest_t *T, np.PyArrayObject *q) except NULL
    np.PyArrayObject *tdigest_cdf_ndarray(tdigest_t *T, np.PyArrayObject *x) except NULL


CENTROID_DTYPE = np.dtype([('mean', np.float64), ('weight', np.float64)])


@cython.boundscheck(False)
cdef inline void _cdf_to_hist(double[:] cdf, double[:]  hist, double size, int hist_size):
    cdef size_t i
    for i in range(hist_size):
        hist[i] = (cdf[i + 1] - cdf[i]) * size


cdef class TDigest:
    """TDigest(compression=100.0)

    An approximate histogram.

    Parameters
    ----------
    compression : float, optional
        The compression factor to use. Larger numbers provide more accurate
        estimates, but use more memory. For most uses, the default should
        suffice.

    Notes
    -----
    This implements the "MergingDigest" variant of the T-Digest algorithm
    descibed in [1]_. The reference java implementation can be found at [2]_.

    References
    ----------
    .. [1] Dunning, Ted, and Otmar Ertl. "Computing Extremely Accurate
       Quantiles Using T-Digests." https://github.com/tdunning/t-digest/blob/
       master/docs/t-digest-paper/histo.pdf

    .. [2] https://github.com/tdunning/t-digest
    """
    cdef tdigest_t *tdigest

    def __cinit__(self, compression=100.0):
        if not isfinite(compression):
            raise ValueError("Compression must be finite")
        self.tdigest = tdigest_new(compression)
        if self.tdigest == NULL:
            raise MemoryError()

    def __dealloc__(self):
        if self.tdigest != NULL:
            tdigest_free(self.tdigest)

    def __repr__(self):
        return ("TDigest<compression={0}, "
                "size={1}>").format(self.compression, self.size())

    @property
    def compression(self):
        """The compression factor for this digest"""
        return self.tdigest.compression

    def min(self):
        """min(self)

        The minimum value in the digest."""
        tdigest_flush(self.tdigest)
        if self.tdigest.total_weight > 0:
            return self.tdigest.min
        return NAN

    def max(self):
        """max(self)

        The maximum value in the digest."""
        tdigest_flush(self.tdigest)
        if self.tdigest.total_weight > 0:
            return self.tdigest.max
        return NAN

    def size(self):
        """size(self)

        The sum of the weights on all centroids."""
        return self.tdigest.total_weight + self.tdigest.buffer_total_weight

    def cdf(self, x):
        """cdf(self, x)

        Compute an estimate of the fraction of all points added to this digest
        which are <= x.

        Parameters
        ----------
        x : array_like or float
        """
        x = np.asarray(x)
        if not np.can_cast(x, 'f8', casting='safe'):
            raise TypeError("x must be numeric")
        if not np.isfinite(x).all():
            raise ValueError("x must be finite")
        out = <np.ndarray>tdigest_cdf_ndarray(self.tdigest, <np.PyArrayObject*>x)
        if out.ndim == 0:
            return np.float64(out)
        return out

    def quantile(self, q):
        """quantile(self, q)

        Compute an estimate of the qth percentile of the data added to
        this digest.

        Parameters
        ----------
        q : array_like or float
            A number between 0 and 1 inclusive.
        """
        q = np.asarray(q)
        if not np.can_cast(q, 'f8', casting='safe'):
            raise TypeError("q must be numeric")
        if not np.isfinite(q).all():
            raise ValueError("q must be finite")
        out = <np.ndarray>tdigest_quantile_ndarray(self.tdigest, <np.PyArrayObject*>q)
        if out.ndim == 0:
            return np.float64(out)
        return out

    def histogram(self, bins=10, range=None):
        """histogram(self, bins=10, range=None)

        Compute a histogram from the digest.

        Parameters
        ----------
        bins : int or array_like, optional
            If ``bins`` is an int, it defines the number of equal width bins in
            the given range. If ``bins`` is an array_like, the values define
            the edges of the bins (rightmost edge inclusive), allowing for
            non-uniform bin widths. The default is 10.

        range : (float, float), optional
            The lower and upper bounds to use when generating bins. If not
            provided, the digest bounds ``(t.min(), t.max())`` are used. Note
            that this option is ignored if the bin edges are provided
            explicitly.

        Returns
        -------
        hist : array
            The values of the histogram.
        bin_edges : array
            The edges of the bins. ``len(bin_edges) == len(hist) + 1``.
        """
        cdef double left = 0
        cdef double right = 0
        cdef double size = self.size()
        if range is None:
            if size != 0:
                left = self.min()
                right = self.max()
        else:
            left = <double?>range[0]
            right = <double?>range[1]
            if not (isfinite(left) and isfinite(right)):
                raise ValueError("range parameters must be finite")
            elif right < left:
                raise ValueError("max must be larger than min for range "
                                 "parameter")
        if right == left:
            left -= 0.5
            right += 0.5

        if isinstance(bins, int):
            if bins < 1:
                raise ValueError("bins must be >= 1")
            bin_edges = np.linspace(left, right, bins + 1, endpoint=True)
        else:
            bin_edges = np.asarray(bins).astype('f8', copy=False)
            if bin_edges.ndim != 1:
                raise ValueError("bins must be a 1-dimensional array")
            elif not np.isfinite(bins).all():
                raise ValueError("bins must be finite")
            elif (np.diff(bin_edges) < 0).any():
                raise ValueError("bins must increase monotonically")

        cdef int hist_size = bin_edges.size - 1
        cdef np.ndarray[double, ndim=1] hist = np.zeros(hist_size,
                                                        dtype=np.float64)
        if size != 0:
            _cdf_to_hist(self.cdf(bin_edges), hist, size, hist_size)

        return hist, bin_edges

    def centroids(self):
        """centroids(self)

        Returns a numpy array of all the centroids in the digest. Note that
        this array is a *copy* of the internal data.
        """
        cdef size_t n
        tdigest_flush(self.tdigest)
        n = self.tdigest.ncentroids

        cdef np.ndarray[centroid_t, ndim=1] result = np.empty(n, dtype=CENTROID_DTYPE)
        if n > 0:
            memcpy(result.data, self.tdigest.centroids, n * sizeof(centroid_t))
        return result

    def __reduce__(self):
        return (TDigest, (self.compression,), self.__getstate__())

    def __getstate__(self):
        return (self.centroids(), self.tdigest.total_weight,
                self.tdigest.min, self.tdigest.max)

    def __setstate__(self, state):
        self.tdigest.total_weight = <double>state[1]
        self.tdigest.min = <double>state[2]
        self.tdigest.max = <double>state[3]

        cdef np.ndarray centroids = state[0]
        cdef int n = len(centroids)
        if n > 0:
            memcpy(self.tdigest.centroids, centroids.data,
                   n * sizeof(centroid_t))
            self.tdigest.ncentroids = n

    def add(self, double x, double w=1):
        """add(self, x, w=1)

        Add a sample to this digest.

        Parameters
        ----------
        x : float
            The value to add.
        w : float, optional
            The weight of the value to add. Default is 1.
        """
        # Don't check w in the common case where w is 1
        if w != 1 and (not isfinite(w) or w <= 0):
            raise ValueError("w must be > 0 and finite")
        tdigest_add(self.tdigest, x, w)

    def update(self, x, w=1):
        """update(self, x, w=1)

        Add many samples to this digest.

        Parameters
        ----------
        x : array_like
            The values to add.
        w : array_like, optional
            The weight (or weights) of the values to add. Default is 1.
        """
        x = np.asarray(x)
        check_w = not (np.isscalar(w) and w == 1)
        w = np.asarray(w)

        if not np.can_cast(x, 'f8', casting='safe'):
            raise TypeError("x must be numeric")

        if not np.can_cast(w, 'f8', casting='safe'):
            raise TypeError("w must be numeric")

        if check_w and (not np.isfinite(w).all() or (w <= 0).any()):
            raise ValueError("w must be > 0 and finite")

        tdigest_update_ndarray(self.tdigest, <np.PyArrayObject*>x,
                               <np.PyArrayObject*>w)

    def merge(self, *args):
        """merge(self, *args)

        Update this digest inplace with data from other digests.

        Parameters
        ----------
        args : TDigests
            TDigests to merge into this one.
        """
        if not all(isinstance(i, TDigest) for i in args):
            raise TypeError("All arguments to merge must be TDigests")
        cdef TDigest td
        for td in args:
            tdigest_merge(self.tdigest, td.tdigest)

    def scale(self, double factor):
        """scale(self, factor)

        Return a new TDigest, with the weights all scaled by factor.

        Parameters
        ----------
        factor : number
            The factor to scale the weights by. Must be > 0.
        """
        if factor <= 0 or not isfinite(factor):
            raise ValueError("factor must be > 0 and finite")
        cdef TDigest out = copy(self)
        tdigest_scale(out.tdigest, factor)
        return out



import os
import sys

import numpy as np
from Cython import Tempita as tempita
from Cython.Build import cythonize
from setuptools import setup
from setuptools.extension import Extension

import versioneer

compile_args = dict(
    library_dirs=[
        os.path.abspath(os.path.join(np.get_include(), "..", "lib")),
    ],
    include_dirs=[np.get_include(), "crick/klib"],
    libraries=["npymath"],
)

if "--debug" in sys.argv:
    sys.argv.remove("--debug")
    compile_args["undef_macros"] = ["NDEBUG"]
    compile_args["extra_compile_args"] = ["-O0"]


def generate_code(templates):
    """Generate code from template files"""
    for template in templates:
        # template extention must be .in
        assert template.endswith(".in")
        outfile = template[:-3]

        if os.path.exists(outfile) and (
            not os.path.exists(template)
            or os.stat(template).st_mtime < os.stat(outfile).st_mtime
        ):
            # If output is present but template isn't, or if template is not
            # updated no need to generate
            continue

        with open(template, "r") as f:
            tmpl = f.read()
        output = tempita.sub(tmpl)

        with open(outfile, "w") as f:
            f.write(output)


templates = ["crick/space_saving_stubs.c.in"]

generate_code(templates)

extensions = [
    Extension("crick.tdigest", ["crick/tdigest.pyx"], **compile_args),
    Extension("crick.space_saving", ["crick/space_saving.pyx"], **compile_args),
    Extension("crick.stats", ["crick/stats.pyx"], **compile_args),
]

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=cythonize(extensions),
)



cimport cython
cimport numpy as np
from cpython.ref cimport Py_INCREF, PyObject

import numpy as np

np.import_array()


cdef extern from "space_saving_stubs.c":
    np.int64_t asint64(np.float64_t key)

    ctypedef struct spsv_t:
        np.intp_t capacity
        np.intp_t size
        np.intp_t head

    # int64
    ctypedef struct counter_int64_t:
        np.int64_t item
        np.int64_t count
        np.int64_t error

    ctypedef struct node_int64_t:
        np.intp_t next
        np.intp_t prev
        counter_int64_t counter

    ctypedef struct spsv_int64_t:
        np.intp_t size
        np.intp_t head
        node_int64_t *list

    spsv_int64_t *spsv_int64_new(int capacity)
    void spsv_int64_free(spsv_int64_t *T)
    int spsv_int64_add(spsv_int64_t *T, np.int64_t item,
                       np.int64_t count) except -1
    int spsv_int64_merge(spsv_int64_t *T1, spsv_int64_t *T2) except -1
    int spsv_int64_set_state(spsv_int64_t *T, counter_int64_t *counters,
                             np.intp_t size) except -1
    int spsv_int64_update_ndarray(spsv_int64_t *T, np.PyArrayObject *item,
                                  np.PyArrayObject *count) except -1

    # object
    ctypedef struct counter_object_t:
        PyObject *item
        np.int64_t count
        np.int64_t error

    ctypedef struct node_object_t:
        np.intp_t next
        np.intp_t prev
        counter_object_t counter

    ctypedef struct spsv_object_t:
        np.intp_t size
        np.intp_t head
        node_object_t *list

    spsv_object_t *spsv_object_new(int capacity)
    void spsv_object_free(spsv_object_t *T)
    int spsv_object_add(spsv_object_t *T, object item,
                        np.int64_t count) except -1
    int spsv_object_merge(spsv_object_t *T1, spsv_object_t *T2) except -1
    int spsv_object_set_state(spsv_object_t *T, counter_object_t *counters,
                              np.intp_t size) except -1
    int spsv_object_update_ndarray(spsv_object_t *T, np.PyArrayObject *item,
                                   np.PyArrayObject *count) except -1


# Repeat struct definition for numpy
_int64_offsets = [<size_t> &(<counter_int64_t*> NULL).item,
                  <size_t> &(<counter_int64_t*> NULL).count,
                  <size_t> &(<counter_int64_t*> NULL).error]

cdef np.dtype COUNTER_INT64_DTYPE = np.dtype(
        {'names': ['item', 'count', 'error'],
         'formats': [np.int64, np.int64, np.int64],
         'offsets': _int64_offsets})

cdef np.dtype COUNTER_FLOAT64_DTYPE = np.dtype(
        {'names': ['item', 'count', 'error'],
         'formats': [np.float64, np.int64, np.int64],
         'offsets': _int64_offsets})

cdef np.dtype COUNTER_OBJECT_DTYPE = np.dtype(
        {'names': ['item', 'count', 'error'],
         'formats': [object, np.int64, np.int64],
         'offsets': [<size_t> &(<counter_object_t*> NULL).item,
                     <size_t> &(<counter_object_t*> NULL).count,
                     <size_t> &(<counter_object_t*> NULL).error
                     ]
         })


class TopKResult(tuple):
    """TopKResult(item, count, error)

    A result from ``SpaceSaving.topk``.

    Attributes
    ----------
    item
        The item, matches the dtype from ``SpaceSaving``.
    count : int
        The estimated frequency.
    error : int
        An upper bound on the error for ``count``. The number of occurrences of
        ``item`` is guaranteed to be in ``count <= actual <= count + error``.
    """

    __slots__ = ()

    def __new__(_cls, item, count, error):
        return tuple.__new__(_cls, (item, count, error))

    def __repr__(self):
        return 'TopKResult(item=%r, count=%r, error=%r)' % self

    def __getnewargs__(self):
        return tuple(self)

    def __dict__(self):
        return dict(zip(('item', 'count', 'error'), self))

    def __getstate__(self):
        pass

    @property
    def item(self):
        return self[0]

    @property
    def count(self):
        return self[1]

    @property
    def error(self):
        return self[2]


cdef class SpaceSaving:
    """SpaceSaving(capacity=20, dtype='f8')

    Approximate TopK.

    Keeps a summary of ``capacity`` counters, for tracking the most frequent
    elements in a stream.

    Parameters
    ----------
    capacity : int, optional
        The number of items to track. A larger number gives higher accuracy,
        but uses more memory. Default is 20.
    dtype : dtype, optional
        The dtype of the objects to track. Currently only ``float64``,
        ``int64``, and ``object`` are supported. Other integer/float types will
        be upcast to match. Default is ``float64``.

    Notes
    -----
    This implements the "Space-Saving" algorithm using the "Stream-Summary"
    datastructure described in [1]_. The merge algorithm is taken from [2]_.

    References
    ----------
    .. [1] Metwally, Ahmed, Divyakant Agrawal, and Amr El Abbadi. "Efficient
       computation of frequent and top-k elements in data streams."
       International Conference on Database Theory. Springer Berlin Heidelberg,
       2005.

    .. [2] Cafaro, Massimo, Marco Pulimeno, and Piergiulio Tempesta. "A
       parallel space saving algorithm for frequent items and the Hurwitz zeta
       distribution." Information Sciences 329 (2016): 1-19.
    """
    cdef spsv_t *summary
    cdef readonly np.dtype dtype

    def __cinit__(self, int capacity=20, dtype='f8'):
        if capacity <= 0:
            raise ValueError("capacity must be > 0")

        dtype = np.dtype(dtype)

        if np.PyDataType_ISOBJECT(dtype):
            self.summary = <spsv_t *>spsv_object_new(capacity)
        elif np.PyDataType_ISINTEGER(dtype):
            dtype = np.dtype(np.int64)
            self.summary = <spsv_t *>spsv_int64_new(capacity)
        elif np.PyDataType_ISFLOAT(dtype):
            dtype = np.dtype(np.float64)
            self.summary = <spsv_t *>spsv_int64_new(capacity)
        else:
            raise ValueError("dtype %s not supported" % dtype)
        self.dtype = dtype

        if self.summary == NULL:
            raise MemoryError()

    def __dealloc__(self):
        if self.summary != NULL:
            if np.PyDataType_ISOBJECT(self.dtype):
                spsv_object_free(<spsv_object_t *>self.summary)
            else:
                spsv_int64_free(<spsv_int64_t *>self.summary)

    def __repr__(self):
        return ("SpaceSaving<capacity={0}, dtype={1}, "
                "size={2}>").format(self.capacity, self.dtype, self.size())

    @property
    def capacity(self):
        return self.summary.capacity

    cpdef int size(self):
        """size(self)

        The number of active counters."""
        return self.summary.size

    def counters(self):
        """counters(self)

        Returns a numpy array of all the counters in the summary. Note that
        this array is a *copy* of the internal data.
        """
        return self.topk(self.size())

    def __reduce__(self):
        return (SpaceSaving, (self.capacity, self.dtype), self.__getstate__())

    def __getstate__(self):
        return (self.counters(),)

    def __setstate__(self, state):
        cdef np.ndarray counters = state[0]
        cdef np.intp_t size = len(counters)
        if np.PyDataType_ISOBJECT(self.dtype):
            spsv_object_set_state(<spsv_object_t*>self.summary,
                                  <counter_object_t*>counters.data, size)
        else:
            if np.PyDataType_ISFLOAT(self.dtype):
                counters = counters.view(COUNTER_INT64_DTYPE)
            spsv_int64_set_state(<spsv_int64_t*>self.summary,
                                 <counter_int64_t*>counters.data, size)

    def add(self, item, int count=1):
        """add(self, item, count=1)

        Add an element to the summary.

        Parameters
        ----------
        item
            The item to add. If it doesn't match the summary dtype, conversion
            will be attempted.
        count : int, optional
            The count of the item to add. Default is 1.
        """
        if count <= 0:
            raise ValueError("count must be > 0")

        if np.PyDataType_ISOBJECT(self.dtype):
            spsv_object_add(<spsv_object_t *>self.summary, item, count)
        elif np.PyDataType_ISFLOAT(self.dtype):
            spsv_int64_add(<spsv_int64_t *>self.summary, asint64(item), count)
        else:
            spsv_int64_add(<spsv_int64_t *>self.summary, item, count)

    def update(self, item, count=1):
        """update(self, item, count=1)

        Add many elements to the summary.

        Parameters
        ----------
        item : array_like
            The item to add. If they don't match the summary dtype, conversion
            will be attempted.
        count : array_like, optional
            The count (or counts) of the item to add. Default is 1.
        """
        item = np.asarray(item)
        check_count = not (np.isscalar(count) and count == 1)
        count = np.asarray(count)

        item = item.astype(self.dtype, casting='safe', copy=False)
        count = count.astype(np.int64, casting='safe', copy=False)

        if check_count and (count <= 0).any():
            raise ValueError("count must be > 0")

        if np.PyDataType_ISOBJECT(self.dtype):
            spsv_object_update_ndarray(<spsv_object_t *>self.summary,
                                       <np.PyArrayObject*>item,
                                       <np.PyArrayObject*>count)
        else:
            if np.PyDataType_ISFLOAT(self.dtype):
                item = item.view('i8')
            spsv_int64_update_ndarray(<spsv_int64_t *>self.summary,
                                      <np.PyArrayObject*>item,
                                      <np.PyArrayObject*>count)

    cpdef topk(self, int k, astuples=False):
        """topk(self, k, astuples=False)

        Estimate the top k elements.

        Parameters
        ----------
        k : int
            The number of most frequent elements to estimate.
        astuples : bool, optional
            If False [default], the result is a numpy record array, otherwise
            it's a list of ``TopKResult`` tuples.

        Returns
        -------
        topk : array or list of TopKResult
            Returns a record array, or a list of ``TopKResult`` tuples. In both
            cases, the result has the following fields:

            - item: The item
            - count: The estimated frequency.
            - error: An upper bound on the error for ``count``. The number of
                     occurrences of ``item`` is guaranteed to be in
                     ``count <= actual <= count + error``.
        """
        cdef np.ndarray out
        if k < 0:
            raise ValueError("k must be >= 0")

        if np.PyDataType_ISOBJECT(self.dtype):
            out = object_counters(<spsv_object_t*>self.summary, k)
        else:
            out = int64_counters(<spsv_int64_t*>self.summary, k)
            if np.PyDataType_ISFLOAT(self.dtype):
                out = out.view(COUNTER_FLOAT64_DTYPE)
        if astuples:
            return [TopKResult(*i) for i in out]
        return out

    def merge(self, *args):
        """merge(self, *args)

        Update this summary in-place with data from other summaries.

        Parameters
        ----------
        args : SpaceSaving(s)
            SpaceSaving(s) to merge into this one.
        """
        if not all(isinstance(i, SpaceSaving) for i in args):
            raise TypeError("All arguments to merge must be SpaceSavings")
        if not all(i.dtype == self.dtype for i in args):
            raise ValueError("All arguments to merge must have same dtype")
        cdef SpaceSaving s
        if np.PyDataType_ISOBJECT(self.dtype):
            for s in args:
                spsv_object_merge(<spsv_object_t*>self.summary,
                                  <spsv_object_t*>s.summary)
        else:
            for s in args:
                spsv_int64_merge(<spsv_int64_t*>self.summary,
                                 <spsv_int64_t*>s.summary)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int64_counters(spsv_int64_t *summary, int k):
    cdef:
        int i = summary.head
        int count
        node_int64_t c
        np.ndarray[counter_int64_t, ndim=1] out

    k = min(summary.size, k)
    out = np.empty(k, dtype=COUNTER_INT64_DTYPE)
    for count in range(k):
        c = summary.list[i]
        out[count] = c.counter
        i = c.next

    return out


@cython.boundscheck(False)
@cython.wraparound(False)
cdef object_counters(spsv_object_t *summary, int k):
    cdef:
        int i = summary.head
        int count
        node_object_t c
        np.ndarray out
        counter_object_t *data

    k = min(summary.size, k)
    out = np.empty(k, dtype=COUNTER_OBJECT_DTYPE)
    data = <counter_object_t *>out.data

    for count in range(k):
        c = summary.list[i]
        Py_INCREF(<object>c.counter.item)
        data[count] = c.counter
        i = c.next
    return out
